


#ifndef XG_C
#define XG_C


#include <stdint.h>
#include <unistd.h> // size_t
#ifndef NOCODE
#define NOCODE
#include "helper.c"
#undef NOCODE

#define SAFER 1
#endif // #ifndef NOCODE


	// The GSM 7bit-8bit-embedding is odd:
	// src= {0xC8, 0xF2, 0x1A}
	//    = 11001000 11110010 00011010
	//    = tgfedcba BAzyxwvu 000GFEDC
	//    =  gfedcba  zyxwvut  GFEDCBA
	//    = o1001000 o1100101 o1101011 = 0x48 0x65 0x6b 
	//    = "Hel"
	//    = dst


#ifndef NOCODE
// accessnum==0 -> use 0x7F
// accessnum==1 -> use 0x80 0x3F
// accessnum==2 -> use      0xC0 0x1F
// accessnum==3 -> use           0xE0 0x0F
// accessnum==4 -> use                0xF0 0x07
// accessnum==5 -> use                     0xF8 0x03
// accessnum==6 -> use                          0xFC 0x01
// accessnum==7 -> use                               0xFE
static int8_t get7bit(uint8_t src[2], uint8_t accessnum) {
	if (!accessnum) return src[0] & 0x7F;
	if (accessnum > 7) return -1;
	
	uint16_t swapped = src[1];
	swapped <<= 8;
	swapped += src[0];
	swapped >>= (8-accessnum);
	return (int16_t)(swapped&0x007F);
}
#endif // #ifndef NOCODE



// write: pointer to be written to. This will be 7-bit plaintext output (for typical input).
// read: pointer to be read from. Hex string (upper/lower case) is expected.
//   Allowed: read == write
//   Not allowed: ( (write) < read && read < (write+size))
// size: number of input bytes. Must be a multiple of two.
// returns number of output bytes (NOT null terminated!).
//     0.5*size <= [Number of output bytes] < 1.0*size
// returns negative number on error.
int pdu7bit2txt(char* write, char* read, size_t size)
#ifndef NOCODE
{
	uint8_t accessnum = 0;
	if (!size)  return 0;
	if (size&1) return -1;
	if (read > write && read < (write+size)) return -9;
	
	uint8_t nextbyte = hexpair2int(read);
	int written = 0;
	size -= 2; read += 2;
	do {
		uint8_t thisbyte = nextbyte;
		nextbyte = 0;
		if (size) {
			nextbyte = hexpair2int(read);
			size -= 2; read += 2;
		}
		
		if (!accessnum) {
			*write++ = get7bit((uint8_t[]){thisbyte, 0}, 0);
			accessnum++;
		}
		
		*write++ = get7bit((uint8_t[]){thisbyte, nextbyte}, accessnum);
		accessnum++;
		if (accessnum & 8) {
			written += 8;
			accessnum = 0;
		}
	}
	while(size || accessnum == 7);
	written += accessnum;
	return written;
}
#endif // #ifndef NOCODE
;




// write: pointer to be written to. This will be a upper case hex string.
// read: pointer to be read from. Must be 7-bit plaintext input.
//   Allowed: read == write
//   Not allowed: ( (write) < read && read < (write+size))

// returns number of output characters (NOT null terminated!).
//     1.0*size <= [Number of output bytes] < 2.0*size
//     precisely:  [Number of output bytes] = ceil(size*2*7/8) = floor((size*2*7+7)/8)
// returns negative number on error.
int txt2pdu7bit(char* write, char* read, size_t size)
#ifndef NOCODE
{
	// It is paradox that "deflating" the data increases
	// size, because of hex representation.
	
	
	if (read > write && read < (write+size)) return -9;
	
	// Operate from back to front
	read = &(read[size]);
	int full_7x2_bunches = size/8;
	write = &(write[full_7x2_bunches*2*7]); // write position
	
	int remainder = size % 8;
	int accessnum = remainder; // different name, same reality
	write += accessnum*2;
	int retW = full_7x2_bunches*2*7 + accessnum*2;
	         //= size/8*2*7 + size%8;
	
	// Outer loop: bunches of 7*2B
		// comparing apples to bananans, but its just to test whether at least on of them is > 0
	while(full_7x2_bunches + accessnum) {
		
		if (accessnum == 0) {
			full_7x2_bunches--;
			accessnum = 8;
		}
		
		uint8_t ch0=0, ch1=0;
		do {
			accessnum--;
			uint8_t src = *--read;
#if defined SAFER
			src &= 0x7f;
#endif
			ch0  = src<<(8-accessnum);
			ch1 |= src>>accessnum;
			
			
			if (accessnum != 7) {
				*--write = int2hexchar(ch1&0xf);
				*--write = int2hexchar(ch1>>4);
			}
			ch1 = ch0;
		}
		while(accessnum);
	}
	
	
	return retW;
}
#endif // #ifndef NOCODE
;





#endif // #ifndef XG_C
